from typing import Any
from jupyter_client import KernelManager, BlockingKernelClient, kernelspec
import uuid
import subprocess
import json
import os
import shutil
from pathlib import Path

ERYS_KERNEL_NAME = "erys_kernel_"
ERYS_DISPLAY_NAME = "erys_kernel"

class NotebookKernel:
    """Class for a kernel for each notebook. Contains kernel manager and client used to
    execute code.
    """

    def __init__(self) -> None:
        self.ksm = kernelspec.KernelSpecManager()
        self.venv_path = os.getenv("VIRTUAL_ENV")
        self.in_venv = self.venv_path is not None
        self.kernel_client: BlockingKernelClient = None
        self.kernel_manager: KernelManager = None

        if self.in_venv:
            if self._check_for_ipykernel():
                # only attempt to connect to the kernel if ipykernel is installed
                self.kernel_path = Path(self.venv_path).joinpath("share/jupyter/")
                os.environ["JUPYTER_PATH"] = str(self.kernel_path)
                self.initialized = True
                self.initialize()
            else:
                self.initialized = False
        else:
            self.initialized = False

    @property
    def display_name(self) -> str:
        """Return the kernel name."""
        try:
            return self.kernel_manager.kernel_spec.display_name
        except:
            return ""

    def initialize(self) -> None:
        """Initializes the notebook kernel's kernel manager and kernel client.
        If kernel specs made by `Erys` are found, they are prioritized. 
        """
        
        kernel_spec, kernel_name = self._get_target_kernel_spec() 

        if not kernel_spec:
            kernel_spec, kernel_name = self._create_new_kernel_spec()

        self.connect_to_kernel(kernel_spec, kernel_name)

    def connect_to_kernel_by_name(self, kernel_name: str) -> None:
        """Connect with a kernel with the given name and start a `BlockingKernelClient`
        to use for code executation.

        Args:
            kernel_name: name of the kernel to connect to.
        """
        self.shutdown_kernel() # will shutdown existing client channels and kernel manager

        kernel_spec = self._get_target_kernel_spec(kernel_name=kernel_name)

        if kernel_spec:
            self.connect_to_kernel(kernel_spec, kernel_name) 

    def connect_to_kernel(self, kernel_spec: dict[str, Any], kernel_name) -> None:
        """Connect with a kernel defined by the given kernel spec nad kernel name.
        Then start a `BlockingKernelClient` to use for code execution.

        Args:
            kernel_spec: spec for the kernel to connect to.
            kernel_name: name of the kernel
        """
        self.shutdown_kernel() # will shutdown existing client channels and kernel manager
        self.kernel_manager: KernelManager = KernelManager(kernel_name=kernel_name)  # kernel manager

        # need to manually provide kernel command so that it is not over ridden by 
        # implementation
        self.kernel_manager.kernel_cmd = kernel_spec["argv"]
        self.kernel_manager.kernel_spec.argv = kernel_spec["argv"]
        self.kernel_manager.kernel_spec.language = kernel_spec["language"]
        self.kernel_manager.kernel_spec.display_name = kernel_spec["display_name"]
        self.kernel_manager.kernel_spec.env = kernel_spec["env"]
        self.kernel_manager.kernel_spec.interrupt_mode = kernel_spec["interrupt_mode"]
        self.kernel_manager.kernel_spec.metadata = kernel_spec["metadata"]

        self.kernel_manager.start_kernel()

        self.kernel_client: BlockingKernelClient = self.kernel_manager.client() # kernel client
        self.kernel_client.start_channels()

    def _create_new_kernel_spec(self) -> tuple[dict[str, Any], str]:
        """Creates new kernel spec in the current python environment.

        Returns the kernel spec and kernel name.
        """
        kernel_name = ERYS_KERNEL_NAME + self._generate_id()
        kernel_spec = self._install_custom_kernel(
            kernel_name=kernel_name,
            display_name=ERYS_DISPLAY_NAME,
        )
        return kernel_spec, kernel_name

    def _generate_id(self) -> str:
        """Generate unique id to use in kernel names created by Erys to avoid collision.

        Returns a uuid hex. 
        """
        return uuid.uuid4().hex[:5]

    def _install_custom_kernel(self, kernel_name: str, display_name: str) -> dict[str, Any]:
        """Writes the kernel specs for a custom `erys` kernel to the virtual enrivonments
        jupyter path.

        Returns the created kernel spec.
        """
        spec_path = self.kernel_path.joinpath(f"kernels/{kernel_name}")
        argv = [ 
            str(Path(self.venv_path).joinpath("bin/python")),
            '-Xfrozen_modules=off',
            '-m',
            'ipykernel_launcher',
            '-f',
            '{connection_file}',
        ]

        kernel_spec = {
            "argv": argv,
            'env': {},
            'display_name': display_name,
            'language': 'python',
            'interrupt_mode': 'signal',
            'metadata': {'debugger': True},
        }

        spec_path.mkdir(parents=True, exist_ok=True)
        with open(spec_path.joinpath("kernel.json"), "w") as spec_file:
            json.dump(kernel_spec, spec_file)

        return kernel_spec

    def _check_for_ipykernel(self) -> bool:
        """Check if the virtual enrivonment has `ipykernel` installed.

        Returns: whether `ipykernel` is installed in environment.
        """

        executable = str(Path(self.venv_path).joinpath("bin/python"))
        result = subprocess.run(
            [executable, "-m", "pip", "freeze"],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            text=True,
            check=True
        )
        return any("ipykernel" == package.split("==")[0] for package in result.stdout.splitlines())

    def _get_available_kernels(self) -> dict[str, str]:
        """Find all the available kernels in the current environment.

        Retuns a dictionary with kernel name as key and resource dir as value.
        """
        return self.ksm.get_all_specs()

    def _update_kernel_cmd(self, kernel_spec: dict[str, Any]) -> dict[str, Any]:
        """Expands the relative python executable path in the kernel cmd of a kernel sepc
        to include path to the environment and disables frozen modules.

        Args:
            kernel_spec: the spec to expand executable path for and disable frozen modules. 
        
        Returns kernel spec.
        """
        kernel_cmd: list[str] = kernel_spec["argv"]
        if kernel_cmd[0] in ["python", "python2", "python3"]:
            kernel_cmd[0] = shutil.which(kernel_cmd[0])
        
        if '-Xfrozen_modules=off' not in kernel_cmd:
            kernel_cmd.insert(1, '-Xfrozen_modules=off')

        return kernel_spec

    def _get_target_kernel_spec(self, kernel_name: str | None = None) -> tuple[dict[str, Any], str]:
        """Goes through all the kernel specs and finds the for the kernel in the current python
        environment and check if it has the provided kernel name if any is.

        Returns: the kernel spec and kernel name.
        """
        kernel_specs = self._get_available_kernels()

        if not kernel_specs: return {}

        target_kernel_spec = {}
        target_kernel_name = ""
        for name, spec in kernel_specs.items():
            resource_dir = spec["resource_dir"]
            if Path(resource_dir).is_relative_to(self.venv_path):
                target_kernel_spec = spec["spec"]
                target_kernel_name = name

                if kernel_name and name == kernel_name: break
                elif kernel_name is None and name.startswith(ERYS_KERNEL_NAME): break
        
        if target_kernel_spec:
            target_kernel_spec = self._update_kernel_cmd(target_kernel_spec)

        return target_kernel_spec, target_kernel_name


    def get_kernel_info(self) -> dict[str, str]:
        """Get the kernel info for the notebook metadata.

        Returns: the dictionary representing the kernel info.
        """
        return {"name": self.kernel_manager.kernel_name}

    def get_kernel_spec(self) -> dict[str, str]:
        """Get the kernel spec for the notebook metadata.

        Returns: the dictionary representing the kernel spec.
        """
        spec = self.kernel_manager.kernel_spec
        if spec:
            return {
                "display_name": spec.display_name,
                "language": spec.language,
                "name": spec.name,
            }

        return {
            "display_name": "",
            "language": "",
            "name": "",
        }

    def get_language_info(self) -> dict[str, Any]:
        """Get the language info for the notebook metadata.

        Returns: the dictionary representing the language info.
        """
        language_info = {}
        try:
            self.kernel_client.kernel_info()
            msg = self.kernel_client.get_shell_msg(timeout=5)

            if msg["header"]["msg_type"] == "kernel_info_reply":
                language_info = msg["content"].get("language_info", {})
        finally:
            return language_info

    def run_code(self, code: str) -> list[dict[str, Any]]:
        """Run provided code string with the kernel. Uses the iopub channel to get results.

        Args:
            code: code string.

        Returns: the outputs of executing the code with the kernel.
        """
        if not self.initialized: return None

        self.kernel_client.execute(code)

        # Read the output from the iopub channel
        outputs = []
        execution_count = None
        while True:
            try:
                msg = self.kernel_client.get_iopub_msg()
                msg_type = msg["header"]["msg_type"]
                match msg_type:
                    case "status":
                        if msg["content"]["execution_state"] == "idle":
                            break
                    case "execute_input":
                        # if no execute output is present for execution, execution count can
                        # be found from the execute_input output
                        execution_count = msg["content"]["execution_count"]
                    case "display_data" | "stream" | "error" | "execute_result":
                        output = msg["content"]
                        output["output_type"] = msg_type
                        outputs.append(output)
            except Exception:
                pass
        return outputs, execution_count

    def interrupt_kernel(self) -> None:
        """Interrupt the kernel."""
        if not self.initialized: return None
        self.kernel_manager.interrupt_kernel()

    def restart_kernel(self) -> None:
        """Restart the kernel."""
        if not self.initialized: return None

        self.kernel_client.stop_channels()
        self.kernel_manager.restart_kernel()
        self.kernel_client: BlockingKernelClient = self.kernel_manager.client()
        self.kernel_client.start_channels()

    def shutdown_kernel(self) -> None:
        """Shutdown the kernel."""
        if self.kernel_client:
            self.kernel_client.stop_channels()

        if self.kernel_manager:
            self.kernel_manager.shutdown_kernel()


if __name__ == "__main__":
    nk = NotebookKernel()
    print(nk.get_kernel_spec())
    print(nk.get_language_info())
    print(nk.get_kernel_info())


from textual.widgets import TextArea, Label, Static
from textual.events import MouseDown, Key, Enter, Leave
from textual.reactive import var
from textual.containers import VerticalGroup
from textual.binding import Binding

from time import time
import pyperclip
from typing import Any
import uuid

DOUBLE_CLICK_INTERVAL = 0.4  # Threshold for interval between mousedown in seconds
COLLAPSED_COLOR = "green"
EXPANDED_COLOR = "white"


# https://github.com/jupyter/enhancement-proposals/blob/master/62-cell-id/cell-id.md
def get_cell_id(id_length=8):
    """Generate the cell id for cells in notebook."""
    return uuid.uuid4().hex[:id_length]


class StaticBtn(Static):
    """Widget to use as button instead of textual's `Button`."""

    def on_enter(self, event: Enter):
        """Add left border when hovering over."""
        self.styles.border_left = "solid", "gray"

    def on_leave(self, event: Leave):
        """Remove borders when mouse leaves."""
        self.styles.border = None


class CollapseLabel(Label):
    """Custom label to use as the collapse button for a cell."""

    collapsed = var(False, init=False)  # keep track of collapse state

    def __init__(
        self, parent_cell: "Cell", collapsed: bool = False, id: str = ""
    ) -> None:
        super().__init__("\n┃\n┃", id=id)
        self.collapsed = collapsed
        self.parent_cell: Cell = parent_cell
        self.prev_switcher = None

    def on_click(self) -> None:
        """Toggle the collapsed state on clicks."""
        self.collapsed = not self.collapsed

    def watch_collapsed(self, collapsed) -> None:
        """Watched method to switch the content from the input text area for cell
        to the collapsed label.

        Args:
            collapsed: updated collapsed state.
        """
        if collapsed:
            # get the placeholder from the input text and set the collapsed display to it
            placeholder = self.get_placeholder(self.parent_cell.input_text.text)
            self.parent_cell.collapsed_display.update(f"{placeholder}...")
            self.prev_switcher = self.parent_cell.switcher.current

            self.parent_cell.switcher.current = "collapsed-display"

            if (
                self.parent_cell.cell_type == "code"
            ):  # if code cell, hide the execution count
                self.parent_cell.exec_count_display.display = False

            self.styles.color = COLLAPSED_COLOR
            self.update("\n┃")
        else:
            # set the switcher to the previous widget it was displaying
            self.parent_cell.switcher.current = self.prev_switcher

            if (
                self.parent_cell.cell_type == "code"
            ):  # if code cell, display the execution count
                self.parent_cell.exec_count_display.display = True

            self.styles.color = EXPANDED_COLOR
            self.update("\n┃\n┃")

    def get_placeholder(self, text: str) -> str:
        """Get the placeholder to represent the cell's text area. Either the first line or an
        empty string.

        Args:
            text: the content of the input text

        Returns: placeholder representing the collapsed input text.
        """
        split = text.splitlines()
        if len(split) == 0:
            return ""

        for line in split:
            if line != "":
                return line


class CopyTextArea(TextArea):
    """Widget to contain text that can be copied."""

    def on_key(self, event: Key) -> None:
        """Key event handler to copy selected text to system clipboard when ctrl+c is pressed."""
        match event.key:
            case "ctrl+c":
                pyperclip.copy(self.selected_text)


class SplitTextArea(CopyTextArea):
    """Widget to contain text that can be split."""

    BINDINGS = [
        ("ctrl+backslash", "split_cell", "Split Cell")
    ]

    def on_key(self, event: Key) -> None:
        """Key event handler. Copy selected text to system clipboard. Call escape event handler
        for the parent cell.

        Args:
            event: Key event.
        """
        match event.key:
            case "ctrl+c":
                pyperclip.copy(self.selected_text)
            case "escape":
                cell: Cell = self.parent.parent.parent
                cell.escape(event)

    def action_split_cell(self) -> None:
        """Creates new cell of the same type as parent."""

        # parent cell containing widget
        cell: Cell = self.parent.parent.parent
        string_to_keep = self.get_text_range((0, 0), self.cursor_location)
        string_for_new_cell = self.text[len(string_to_keep) :]

        self.load_text(string_to_keep)

        # create and mount new cell of the same type as we are spliting from
        new_cell = cell.create_cell(string_for_new_cell)
        cell.notebook.cell_container.mount(new_cell, after=cell)
        # connect the new cell to the cells next and before it
        cell.notebook.connect_widget(new_cell)


class Cell(VerticalGroup):
    """Base class for the markdown and code cell."""

    can_focus = True
    _last_click_time: float = 0.0  # keep track of the last mouse click
    merge_select: bool = var(False, init=False)  # whether cell is selected for merging

    # pointers to the next and prevous cells in the notebook
    next = None
    prev = None

    # name of the cell type
    cell_type = ""

    BINDINGS = [
        Binding("c", "collapse", "Collapse Cell", False),
        Binding("ctrl+pageup", "join_above", "Join with Before", False),
        Binding("ctrl+pagedown", "join_below", "Join with After", False),
    ]

    def __init__(
        self,
        notebook,
        source: str,
        language: str,
        metadata: dict[str, Any] = {},
        cell_id: str | None = None,
    ) -> None:
        super().__init__()
        self.notebook = notebook  # notebook cell belongs to
        self.source = source
        self._metadata = metadata
        self._cell_id = cell_id or get_cell_id()
        self.id = f"_{self._cell_id}"
        self._collapsed = metadata.get("collapsed", False)
        self._language = language

        self.collapse_btn = CollapseLabel(
            self, collapsed=self._collapsed, id="collapse-button"
        ).with_tooltip("Collapse")

        self.collapsed_display = Static("", id="collapsed-display")

    async def on_key(self, event: Key) -> None:
        """Key press event handler to 'open' the `Cell` when enter is pressed.

        Args:
            event: Key press event.
        """
        match event.key:
            case "enter":
                await self.open()

    def _on_focus(self):
        """Focus event handler that adds border."""
        if not self.merge_select:
            self.styles.border_left = "solid", "lightblue"
        # self.styles.border = "solid", "lightblue"
        # self.border_subtitle = self._language

    def _on_blur(self):
        """Blur event handler that removes border if not selected for merge."""
        if not self.merge_select:
            self.styles.border = None

    def on_enter(self, event: Enter) -> None:
        """Mouse enter event handler that adds border if not selected for merge and not focused."""
        if self.merge_select:
            return

        if self.notebook.last_focused != self:
            self.styles.border_left = "solid", "grey"

    def on_leave(self, event: Leave) -> None:
        """Mouse leave event handler that removes border if not selected for merge and not focused."""
        if self.merge_select:
            return

        if self.notebook.last_focused != self:
            self.styles.border_left = None

    def on_mouse_down(self, event: MouseDown) -> None:
        """Mouse down event handler. Add cell to the merge list if ctrl is held. If the time
        between mouse down event is below the threshold, call the double click event handler.

        Args:
            event: MouseDown event.
        """
        now = time()
        if event.ctrl:
            if not self.merge_select:
                self.notebook._merge_list.append(self)
            else:
                self.notebook._merge_list.remove(self)

            self.merge_select = not self.merge_select
        elif now - self._last_click_time <= DOUBLE_CLICK_INTERVAL:
            self.on_double_click(event)
        self._last_click_time = now

    def watch_merge_select(self, selected: bool) -> None:
        """Watcher method that updates the border of the cell depending on whether it is selected
        for merging.

        Args:
            selected: whether cell is selected for merging.
        """
        if selected:
            self.styles.border_left = "solid", "yellow"
        else:
            self.styles.border_left = None

    def action_join_above(self) -> None:
        """Merges cell with the previous cell."""
        if self.prev:
            self.prev.merge_cells_with_self([self])

    def action_join_below(self) -> None:
        """Merges cell with the next cell."""
        if self.next:
            self.merge_cells_with_self([self.next])

    def disconnect(self) -> tuple["Cell", str]:
        """Remove self from the linked list of cells. Update the pointers of the surrounding cells
        to point to each other.

        Returns: The next cell to focus on and where the removed cell was relative to it.
        """
        next_focus = None
        position = "after"
        if prev := self.prev:
            next_focus = prev
            position = "after"
            prev.next = self.next

        if next := self.next:
            next_focus = next
            position = "before"
            next.prev = self.prev

        return next_focus, position

    def set_new_id(self) -> None:
        """Update cell id with a new uuid."""
        self._cell_id = get_cell_id()

    def merge_cells_with_self(self, cells) -> None:
        """Merge self with a list of cells by combining content in text areas into self. Should be
        called by the first selected cell in the the cells to merge. The resulting type will be
        self.

        Args:
            cells: List of MarkdownCell | CodeCell to merge with self.
        """
        source = self.input_text.text

        for cell in cells:
            source += "\n"
            source += cell.input_text.text
            cell.disconnect()
            cell.remove()

        self.input_text.load_text(source)
        self.focus()

    def on_double_click(self, event: MouseDown) -> None:
        """Double click event handler."""
        pass

    def escape(self, event: Key) -> None:
        """Event handler for the escape key."""
        raise NotImplementedError()

    @staticmethod
    def from_nb(nb: dict[str, Any], notebook) -> "Cell":
        """Static method to generate a `Cell` from a json/dict that represent a cell."""
        raise NotImplementedError()

    def to_nb(self) -> dict[str, Any]:
        """Static method to generate a `Cell` from a json/dict that represent a cell."""
        raise NotImplementedError()

    def create_cell(self, str: str) -> "Cell":
        """Returns a `CodeCell` with a source. Used for splitting cell."""
        raise NotImplementedError()

    def clone(self, connect: bool = True):
        """Clone a code cell. Used for cut/paste."""
        raise NotImplementedError()



from textual.app import ComposeResult
from textual.widgets import TextArea, Markdown, Static
from textual.reactive import var
from textual.containers import VerticalScroll, Container, HorizontalScroll
from textual.events import Key, DescendantFocus, Click
from textual.binding import Binding

from threading import Lock
from typing import Any
import json
from pathlib import Path
import time

from .markdown_cell import MarkdownCell
from .code_cell import CodeCell, CodeArea, ExecStatus, OutputText, OutputJson, OutputAnsi
from .cell import CopyTextArea, Cell, StaticBtn
from .notebook_kernel import NotebookKernel
from .exec_queue import Queue


MAX_UNDO_LEN = 20
DEFAULT_FILE_NAME = "erys_notebook"


class ButtonRow(HorizontalScroll):
    """Buttton row on top of Notebook"""
    kernel_name: str = var("", init=False)

    def compose(self) -> ComposeResult:
        """Composed with:
        - HorizontalGroup
            - StaticBtn (id=add-code-cell)
            - StaticBtn (id=add-markdown-cell)
            - StaticBtn (id=run-before)
            - StaticBtn (id=run-after)
            - StaticBtn (id=run-all)
            - StaticBtn (id=restart-shell)

        """
        yield StaticBtn("➕ Code", id="add-code-cell")
        yield StaticBtn("➕ Markdown", id="add-markdown-cell")
        yield StaticBtn("▶ Run All", id="run-all")
        yield StaticBtn("▲ Run Before", id="run-before")
        yield StaticBtn("▼ Run After", id="run-after")
        # yield StaticBtn("▶ ↑ Run Before", id="run-before")
        # yield StaticBtn("▶ ↓ Run After", id="run-after")
        yield StaticBtn("🔁 Restart", id="restart-shell")
        self.kernel_label = Static(f"Kernel: {self.kernel_name}", id="kernel-name")
        yield self.kernel_label


    def watch_kernel_name(self, kernel_name: str) -> None:
        """Update the kernel label with the new kernel name.

        Args:
            kernel_name: new kernel name.
        """
        self.kernel_label.update(f"Kernel: {kernel_name}")


class Notebook(Container):
    """Container representing a notebook."""

    BINDINGS = [
        Binding("a", "add_cell_after", "Add Cell After", False),
        Binding("b", "add_cell_before", "Add Cell Before", False),
        Binding("t", "toggle_cell", "Toggle Cell Type", False),
        ("ctrl+d", "delete_cell", "Delete Cell"),
        ("ctrl+u", "undo", "Undo Delete"),
        Binding("ctrl+up", "move_up", "Move Cell Up", False),
        Binding("ctrl+down", "move_down", "Move Cell Down", False),
        Binding("M", "merge_cells", "Merge Cells", False),
        Binding("ctrl+c", "copy_cell", "Copy Cell", False),
        Binding("ctrl+x", "cut_cell", "Cut Cell", False),
        Binding("ctrl+v", "paste_cell", "Paste Cell", False),
        ("ctrl+s", "save", "Save"),
        ("ctrl+w", "save_as", "Save As"),
    ]

    def __init__(self, path: str, id: str, term_app) -> None:
        super().__init__(id=id)

        self.last_focused: Cell | None = None  # keep track of the last focused cell
        self.last_copied: Cell | None = None  # keep track of the copied/cut cell
        self._delete_stack: list[tuple[dict[str, Any], str, str]] = []
        self._merge_list: list[Cell] = []  # list of the cells to be merged.
        self._exec_queue: Queue = Queue()
        self._queue_lock: Lock = Lock()

        self.path = path
        self.notebook_kernel = NotebookKernel()
        self.term_app = term_app

        self._metadata: dict[str, Any] | None = None
        self._nbformat: int = 4
        self._nbformat_minor: int = 5


    def compose(self) -> ComposeResult:
        """Composed with:
        - Container
            - ButtonRow
            - VerticalScroll (id=cell-container)
                - Cell
        """
        self.btn_row = ButtonRow()
        yield self.btn_row
        self.cell_container = VerticalScroll(id="cell-container")
        yield self.cell_container

    def _is_kernel_connected(self) -> bool:
        if not self.notebook_kernel.in_venv:
            self.notify(
                "[bold]Erys[/] was not opened in a virtual environment. Open in a virtual "
                "environment to run notebook.",
                severity="error",
                timeout=10,
            )
        elif not self.notebook_kernel.initialized:
            self.notify(
                "[bold]ipykernel[/] not installed in virtual enrivonment. Install "
                "to run notebook.",
                severity="error",
                timeout=10,
            )

        return self.notebook_kernel.initialized

    def _is_new_notebook(self) -> bool:
        """Checks if notebook is new."""
        if isinstance(self.path, Path): return False
        return self.path[:len(DEFAULT_FILE_NAME)] == DEFAULT_FILE_NAME and self.path[len(DEFAULT_FILE_NAME):].isdigit()

    async def on_mount(self) -> None:
        """Mount event handler that loads a notebook if path is provided."""

        if not self._is_new_notebook():
            self.path = Path(self.path)
            if self.path.exists():
                self.call_after_refresh(self.load_notebook)

        if self._is_kernel_connected():
            self.btn_row.kernel_name = self.notebook_kernel.display_name

        self.call_next(self.focus_file)

        self.run_worker(self.notebook_executor, thread=True)

    def on_unmount(self) -> None:
        """Unmount event handler that shuts down kernel if avaialble."""
        if self.notebook_kernel.initialized:
            self.notebook_kernel.interrupt_kernel()
            time.sleep(0.5)
            self.notebook_kernel.shutdown_kernel()

        self.clear_exec_queue()
        self._exec_queue.enqueue(None)

    def on_descendant_focus(self, event: DescendantFocus) -> None:
        """Descendant focus event handler that assigns the last focused cell depending on what
        widget is being focused on. Chooses the parent cell of the widget being currently focused
        on.

        Args:
            event: descendant focus event.
        """
        # Ignore buttons
        if isinstance(event.widget, CodeCell) or isinstance(event.widget, MarkdownCell):
            self.last_focused = event.widget
        elif any(
            isinstance(event.widget, widgetType)
            for widgetType in [OutputJson, OutputText, OutputAnsi]
        ) or event.widget.id in ["pretty-error-output", "plain-error-output"]:
            self.last_focused = event.widget.parent.parent.parent.parent
        elif any(
            isinstance(event.widget, widgetType)
            for widgetType in [Markdown, CopyTextArea, CodeArea]
        ):
            self.last_focused = event.widget.parent.parent.parent
        elif self.last_focused:
            self.last_focused.focus()

    def on_key(self, event: Key) -> None:
        """Key event handler that
            - disables tab and shift+tab for changing focus,
            - clears merge list when escape is presed
            - moves focus between cells using up and down arrows.

        Args:
            event: key event.
        """
        match event.key:
            case "tab" | "shift+tab":
                event.prevent_default()
                event.stop()
            case "escape":
                for cell in self._merge_list:
                    cell.merge_select = False
                self._merge_list = []
            case "up" if not isinstance(self.app.focused, TextArea):
                if self.last_focused and (prev_cell := self.last_focused.prev):
                    prev_cell.focus()
            case "down" if not isinstance(self.app.focused, TextArea):
                if self.last_focused and (next_cell := self.last_focused.next):
                    next_cell.focus()

    async def on_click(self, event: Click) -> None:
        """Click event handler for button row on top of notebook.

        Args:
            event: click event.
        """
        match event.widget.id:
            case "add-code-cell":
                widget = await self.add_cell(CodeCell, self.last_focused, "after")
                self.call_after_refresh(widget.open)
            case "add-markdown-cell":
                widget = await self.add_cell(MarkdownCell, self.last_focused, "after")
                self.call_after_refresh(widget.open)
            case "restart-shell":
                self.notebook_kernel.restart_kernel()
            case "run-all" | "run-after" | "run-before" if not self.notebook_kernel:
                # if any of the run buttons are pressed check if there is a notebook kernel
                # pass
                if not self._is_kernel_connected():
                    return
            case "run-all":
                await self.run_all_cells()
            case "run-after":
                await self.run_cells_after()
            case "run-before":
                await self.run_cells_before()

    async def action_add_cell_after(self) -> None:
        """Add code cell after current cell."""
        cell = await self.add_cell(CodeCell, self.last_focused, "after")
        cell.focus()

    async def action_add_cell_before(self) -> None:
        """Add code cell before current cell."""
        cell = await self.add_cell(CodeCell, self.last_focused, "before")
        cell.focus()

    def action_save_as(self) -> str:
        """Save notebook as a new file."""

        def check_save_as(path: Path | None) -> None:
            """Callback function to save notebook if save as screen dismisses successfully.

            Args:
                path: string if save as screen dismisses with a file path chosen
                    to save notebook at.
            """
            if path:
                self.path = path
                # notify after successfuly serializing and saving notebook
                self.save_notebook(path)
                self.notify(f"{self.path} saved!")
                # open the saved notebook
                
                # change the tab name
                self.term_app.change_tab_name(self.id, self.path)

        # push the save as screen with the above callback function
        self.app.push_screen("nb_save_as_screen", check_save_as)

    def action_save(self) -> None:
        """Save notebook."""
        # notebooks with path 'new_empty_terminal_notebook' were created by terminal-notebook
        # need save_as call to get file name
        if self._is_new_notebook():
            self.action_save_as()
        else:
            # notify after successfuly serializing and saving notebook
            self.save_notebook(self.path)
            self.notify(f"{self.path} saved!")

    def action_delete_cell(self) -> None:
        """Delete cell."""
        self.delete_cell()

    def action_copy_cell(self) -> None:
        """Copy cell."""
        if not self.last_focused:
            return
        # store serialized representation of copied cell
        self.last_copied = self.last_focused.to_nb()

    def action_cut_cell(self) -> None:
        """Cut cell."""
        if not self.last_focused:
            return

        # store serialized representation of cut cell and delete it
        self.last_copied = self.last_focused.to_nb()
        self.delete_cell()

    async def action_paste_cell(self) -> None:
        """Paste cut/copied cell."""
        if not self.last_copied:
            return

        # generate the `MarkdownCell` or `CodeCell` from the stored serialized copy.
        # remove id and generate new to avoid conflict during copy/paste
        self.last_copied["id"] = None
        match self.last_copied["cell_type"]:
            case "markdown":
                widget = MarkdownCell.from_nb(self.last_copied, self)
            case "code":
                widget = CodeCell.from_nb(self.last_copied, self)

        # mount and connect the widget
        await self.cell_container.mount(widget, after=self.last_focused)
        self.connect_widget(widget)

    async def action_move_up(self) -> None:
        """Move the cell up."""
        if not self.last_focused:
            return

        if self.last_focused.prev:
            # clone the cell
            clone = self.last_focused.clone()
            prev = clone.prev

            # removes the cell from linked list
            prev.next = clone.next
            if clone.next:
                clone.next.prev = prev

            if prev.prev:
                prev.prev.next = clone

            clone.prev = prev.prev
            clone.next = prev
            prev.prev = clone

            # remove the original cell
            await self.last_focused.remove()

            # mount the new cloned cell in the new position
            await self.cell_container.mount(clone, before=clone.next)

            self.last_focused = clone
            self.last_focused.focus()

    async def action_move_down(self) -> None:
        """Move the cell down."""
        if not self.last_focused:
            return

        if self.last_focused.next:
            # clone the cell
            clone = self.last_focused.clone()
            next = clone.next

            # removes the cell
            next.prev = clone.prev
            if clone.prev:
                clone.prev.next = next

            if two_down := next.next:
                two_down.prev = clone

            # adds the cell
            clone.prev = next
            clone.next = next.next
            next.next = clone

            # removes the cell from linked list
            await self.last_focused.remove()

            # mount the new cloned cell in the new position
            await self.cell_container.mount(clone, after=clone.prev)

            self.last_focused = clone
            self.last_focused.focus()

    def action_merge_cells(self) -> None:
        """Merge selected cells by combining content in text areas into the one selected. Should be
        called by the first selected cell in the the cells to merge. The resulting type will be
        of the first selected's type.
        """
        if len(self._merge_list) < 2:
            return
        target: CodeCell | MarkdownCell = self._merge_list[0]
        target.merge_cells_with_self(self._merge_list[1:])
        target.merge_select = False
        self._merge_list = []

    async def action_toggle_cell(self) -> None:
        """Callback for binding to swtich the cell type keeping the input text source."""
        await self.toggle_cell_type()

    def action_undo(self) -> None:
        """Callback for binding to undo deletion."""
        self.undo_delete()

    def undo_delete(self) -> None:
        """Undo last deletion by recreating cell from saved serialized data. Try to find the `Cell`
        with the id stored and mount relative to it with the stored position. If the query by id
        fails, just place relative to the `last_focused` cell. If the stored id is
        None, then widget was the only cell in the notebook when it was deleted.
        """
        if len(self._delete_stack) == 0:
            return

        # get the last deleted cell
        last_delete, position, relative_to_id = self._delete_stack.pop()

        # recreate the deleted cell
        match last_delete["cell_type"]:
            case "markdown":
                widget = MarkdownCell.from_nb(last_delete, self)
            case "code":
                widget = CodeCell.from_nb(last_delete, self)

        if relative_to_id:
            try:
                # attempt to find the cell to mount relative to
                target_widget = self.cell_container.query_one(
                    f"#{relative_to_id}", Cell
                )
            except:
                # if the query fails, the target position should be relative to the last_focused cell
                target_widget = self.last_focused
        else:
            target_widget = None

        # mount it relative to the target_widget
        self.cell_container.mount(widget, **{position: target_widget})

        # update the pointers of the widgets surrounding where it was mounted
        self.connect_widget(widget, target_widget, position)

    def delete_cell(self, remember: bool = True) -> None:
        """Delete a cell and keep track of it in the `_delete_stack` if remember is True.

        Args:
            remember: whether to keep track of the deleted to undo later.
        """
        if not self.last_focused:
            return

        # disconnect the cell from surrounding cells and find new cell to focus on
        next_focus, position = self.last_focused.disconnect()

        if remember:
            # add it to the `delete_stack` for undoing
            id = next_focus.id if next_focus else None
            self._delete_stack.append((self.last_focused.to_nb(), position, id))
            if len(self._delete_stack) > MAX_UNDO_LEN:
                self._delete_stack = self._delete_stack[-MAX_UNDO_LEN:]

        # remove cell
        self.call_after_refresh(self.last_focused.remove)
        self.last_focused = next_focus

        if self.last_focused:
            self.last_focused.focus()

    def clear_exec_queue(self) -> None:
        """Clears execution queue and resets code cells to idle."""
        for cell in self._exec_queue.clear():
            cell.status = ExecStatus.IDLE

    def interrupt_exec(self) -> None:
        """Interrupt execution and clear the execution queue."""
        self.notebook_kernel.interrupt_kernel()
        self.clear_exec_queue()

    async def run_all_cells(self) -> None:
        """Run all code cells."""
        # iterate over all the code cells and run them
        for cell in self.cell_container.children:
            if isinstance(cell, CodeCell):
                cell.status = ExecStatus.QUEUED
                self._exec_queue.enqueue(cell)
        self.last_focused.focus()

    async def run_cells_after(self) -> None:
        """Run code cells after currently focused cell (inclusive)."""
        if not self.last_focused:
            return

        # iterate over all the code cells starting from (including) the current and run them
        cell = self.last_focused
        while cell:
            if isinstance(cell, CodeCell):
                cell.status = ExecStatus.QUEUED
                self._exec_queue.enqueue(cell)
            cell = cell.next
        self.last_focused.focus()

    async def run_cells_before(self) -> None:
        """Run code cells before currently focused cell (not inclusive)."""
        if not self.last_focused:
            return
        # iterate over all the code cells up to (not including) the current and run them
        for cell in self.cell_container.children:
            if cell == self.last_focused:
                break
            if isinstance(cell, CodeCell):
                cell.status = ExecStatus.QUEUED
                self._exec_queue.enqueue(cell)
        self.last_focused.focus()

    async def notebook_executor(self) -> None:
        """Threaded consumer function to run code cells using the `_exec_queue`."""
        while self.notebook_kernel.initialized:
            code_cell: CodeCell = self._exec_queue.dequeue()

            if  code_cell is None: # sentinel for exiting
                break
            src = code_cell.input_text.text

            if src == "":  # only call the kernel execute if there is code
                continue

            code_cell.status = ExecStatus.RUNNING
            outputs, execution_count = self.notebook_kernel.run_code(src)
            code_cell.status = ExecStatus.IDLE
            code_cell.handle_exec_completion(outputs, execution_count)

        return

    async def toggle_cell_type(self) -> None:
        """Swtich the cell type keeping the input text source."""
        if not self.last_focused:
            return

        # get the source in the input text of the cell
        src = self.last_focused.input_text.text
        # get the type to swtich to and construct object
        new_cell_type = (
            MarkdownCell if isinstance(self.last_focused, CodeCell) else CodeCell
        )
        new_cell = new_cell_type(self, src)

        # easiest solution is to mount after the cell that is switching then deleting the original
        await self.cell_container.mount(new_cell, after=self.last_focused)
        self.connect_widget(new_cell)
        self.delete_cell(remember=False)

    def focus_file(self) -> None:
        """Defines what focusing on a notebook does. If there is a cell that was last focused,
        focus on it; otherwise, focus on the `cell_container`.
        """
        if self.last_focused:
            self.call_next(self.last_focused.focus)
        else:
            self.call_next(self.cell_container.focus)

    def save_notebook(self, path: str) -> None:
        """Saves the notebook to the provided path.

        Args:
            path: file path to save notebook at.
        """
        nb = self.to_nb()
        with open(path, "w") as nb_file:
            json.dump(nb, nb_file)

    def load_notebook(self) -> None:
        """Load notebook from a file. Iterate through the cells and generate the `CodeCell` and
        `MarkdownCell` objects from the serialized formats and mount them to the `cell_container`.
        """
        with open(self.path, "r") as notebook_file:
            # if the file is empty, trying to decode json will fail so early exit.
            if self.path.stat().st_size == 0:
                return

            try:
                # if the json decoding fails then the notebook is bad
                content = json.load(notebook_file)
            except json.JSONDecodeError:
                self.notify(f"{self.path} is not a valid notebook. Failed to decode JSON.", severity="error")
                self.term_app.remove_tab(str(self.path))
                return

        for idx, cell in enumerate(content["cells"]):
            if cell["cell_type"] == "code":
                widget = CodeCell.from_nb(cell, self)
            elif cell["cell_type"] == "markdown":
                widget = MarkdownCell.from_nb(cell, self)

            if idx != 0:
                prev.next = widget
                widget.prev = prev
            else:
                self.last_focused = widget

            prev = widget
            self.call_next(self.cell_container.mount, widget)

    async def add_cell(
        self,
        cell_type: CodeCell | MarkdownCell,
        relative_to: CodeCell | MarkdownCell | None,
        position: str = "after",
        **cell_kwargs,
    ) -> CodeCell | MarkdownCell:
        """Add a cell by creating object of `cell_type` with arguments `cell_kwargs` with position
        `position` relative to the widget `relative_to`.

        Args:
            cell_type: the type of the cell being added.
            relative_to: what cell if any it should be added relative to.
            position: where in relation to the focused cell to connect the widget 'after', 'before'.
            **cell_kwargs: key word arguments to use to create the cell.
        """
        kwargs = {position: relative_to}

        widget = cell_type(self, **cell_kwargs)

        await self.cell_container.mount(widget, **kwargs)
        self.connect_widget(widget, position=position)

        return widget

    def connect_widget(
        self, widget: Cell, relative_to: Cell | None = None, position: str = "after"
    ) -> None:
        """Connect the cell (widget) after or before the the `Cell` `relative_to`.

        Args:
            widget: the widget being connected.
            relative_to: relative to which cell the widget is being connected
            position: where in relation to the focused cell to connect the widget 'after', 'before'.
        """
        relative_to = self.last_focused if not relative_to else relative_to

        if (
            not relative_to
        ):  # if no cell has been focused on, set the new cell as the focused
            self.last_focused = widget
            self.last_focused.focus()

        # if positioning after last focused, add widget between last_focused and last_focused.next
        elif position == "after":
            next = relative_to.next
            relative_to.next = widget
            widget.next = next
            widget.prev = relative_to

            if next:
                next.prev = widget

        # if positioning before last focused, add widget between last_focused and last_focused.prev
        elif position == "before":
            prev = relative_to.prev
            relative_to.prev = widget
            widget.next = relative_to
            widget.prev = prev

            if prev:
                prev.next = widget

    def to_nb(self) -> dict[str, Any]:
        """Serialize the `Notebook` to notebook format. Format for notebook is:
            {
                "metadata" : {
                    "signature": "hex-digest",
                    "kernel_info": {
                        "name" : "the name of the kernel"
                    },
                    "language_info": {
                        "name" : "the programming language of the kernel",
                        "version": "the version of the language",
                        "codemirror_mode": "The name of the codemirror mode to use [optional]"
                    }
                },
                "nbformat": 4,
                "nbformat_minor": 0,
                "cells" : [
                ],
            }

        Returns: serialized notebook in notebook format.
        """
        kernel_spec = self.notebook_kernel.get_kernel_spec()
        kernel_info = self.notebook_kernel.get_kernel_info()
        language_info = self.notebook_kernel.get_language_info()

        cells = [cell.to_nb() for cell in self.cell_container.children]

        return {
            "metadata": {
                "kernel_info": kernel_info,
                "kernel_spec": kernel_spec,
                "language_info": language_info,
            },
            "nbformat": self._nbformat,
            "nbformat_minor": self._nbformat_minor,
            "cells": cells,
        }



from collections import deque
from typing import Any
from threading import Condition

class Queue:
    """Custom thread safe queue to use for producer/consumer code cell execution."""

    def __init__(self) -> None:
        self.queue = deque()
        self.condition = Condition()

    def enqueue(self, val: Any) -> None:
        """Enqueue element to the queue.

        Args:
            val: element to enqueue
        """
        with self.condition:
            self.queue.append(val)
            self.condition.notify()

    def dequeue(self) -> Any:
        """Dequeue element from the queue.

        Returns dequeued element.
        """
        with self.condition:
            if not self.queue:
                self.condition.wait()
            return self.queue.popleft()

    def clear(self) -> list[Any]:
        """Clear queue and return previous elements.
        
        Returns elements of the queue before clearing.
        """
        with self.condition:
            elements = list(self.queue)
            self.queue.clear()
            return elements

    def empty(self) -> bool:
        """Returns True if the queue is empty."""
        with self.condition:
            return not self.queue

    def push_left(self, val: Any) -> None:
        """Add element to the begining of the queue.
        
        Args:
            val: element to add.
        """
        with self.condition:
            self.queue.appendleft(val)


from textual.app import ComposeResult
from textual.widgets import Static, DirectoryTree, Input
from textual.containers import Vertical
from textual.screen import Screen
from textual.events import Key, ScreenResume
from textual.validation import ValidationResult, Validator

from pathlib import Path
from typing import Iterable


class NotebookName(Validator):
    """Validator for notebook name."""

    def validate(self, value: str) -> ValidationResult:
        """Checks if the provided name for a noteboook has correct ipynb extension.

        Args:
            value: provided file name for notebook.

        Returns: validation result based on whether extension is correct.
        """
        ext = Path(value).suffix
        if ext == ".ipynb":
            return self.success()
        else:
            return self.failure("File extension is not .ipynb.")


class FilteredDirectoryTree(DirectoryTree):
    """Filtered directory tree containing only directories and notebook files."""

    def filter_paths(self, paths: Iterable[Path]) -> Iterable[Path]:
        return [path for path in paths if path.suffix == ".ipynb" or path.is_dir()]


class SaveAsScreen(Screen[str | None]):
    def __init__(self, is_notebook: bool) -> None:
        super().__init__()
        self.is_notebook = is_notebook
        if is_notebook:
            # input for file name
            self.input = Input(
                placeholder="File Name", id="save-as-input", validators=[NotebookName()]
            )
            # to help select the directory where the file can be saved at
            self.dir_tree = FilteredDirectoryTree(Path.cwd(), id="save-as-dir-tree")
        else:
            self.input = Input(
                placeholder="File Name", id="save-as-input",
            )
            self.dir_tree = DirectoryTree(Path.cwd(), id="save-as-dir-tree")

    def compose(self) -> ComposeResult:
        """Composed with:
        - Screen
            - Vertical (id=save-as)
                - Input (id=save-as-input)
                - Static (id=save-as-dir)
                - FilteredDirectoryTree (id=save-as-dir-tree)
        """
        with Vertical(id="save-as"):
            # displays the directory where the file will be saved at
            self.cur_dir = Static(f"Saving at: {Path.cwd()}", id="save-as-dir")

            yield self.cur_dir
            yield self.input
            yield self.dir_tree

    def on_mount(self) -> None:
        """Properly set directory tree path after mount."""
        self.cur_dir.update(f"Saving at: {Path.cwd()}")
        self.dir_tree.path = Path.cwd()

    def on_screen_resume(self, event: ScreenResume) -> None:
        """Screen resume event handler that resets the path for directory tree and static to pwd.

        Args:
            event: screen resume event.
        """
        self.cur_dir.update(f"Saving at: {Path.cwd()}")
        self.dir_tree.path = Path.cwd()

    def on_directory_tree_file_selected(
        self, event: DirectoryTree.FileSelected
    ) -> None:
        """File selected event handler that sets the file name for saving file to
        the name of the selected file and updates the directory where the file will be saved.

        Args:
            event: file selected event.
        """
        file_dir = event.path.parent
        self.cur_dir.update(f"Saving at: {file_dir}")

        file_name = event.path.name
        # clear and update the input widget to show selected file's file name
        self.input.clear()
        self.input.insert(file_name, 0)
        # don't propagate this event to other handlers
        event.stop()

    def on_directory_tree_directory_selected(
        self, event: DirectoryTree.DirectorySelected
    ) -> None:
        """Directory selected event handler that sets the directory where
        file will be saved to the selected directory.

        Args:
            event: directory selected event.
        """
        self.dir_tree.path = event.path
        self.cur_dir.update(f"Saving at: {event.path}")
        # don't propagate this event to other handlers
        event.stop()

    def on_key(self, event: Key) -> None:
        """Key press event handler that
            - prevents event propagation for the escape, n, ctrl+k, ctrl+l, d keys
            - used backspace for navigating up a directory

        Args:
            event: key event.
        """
        match event.key:
            case "escape":
                self.dismiss(None)
                # don't propagate this event to other handlers
                event.stop()
            case "backspace" if self.app.focused == self.dir_tree:
                parent = Path(self.dir_tree.path).resolve().parent
                self.dir_tree.path = parent
                self.cur_dir.update(f"Saving at: {parent}")
            case "n" | "ctrl+k" | "ctrl+l" | "d":
                # don't propagate this event to other handlers
                event.stop()

    def on_input_submitted(self, event: Input.Submitted) -> None:
        """Input submitted event handler that dismisses the screen with the file path if
        input value is valid; otherwise notifies that the input value is not a valid file name.


        Args:
            event: input submitted event.
        """
        if self.is_notebook:
            if event.validation_result.is_valid:
                file_path = self.dir_tree.path.joinpath(event.value)
                self.dismiss(file_path)
            else:
                self.notify(event.validation_result.failure_descriptions, severity="error")
        else:
            file_path = self.dir_tree.path.joinpath(event.value)
            self.dismiss(file_path)



from __future__ import annotations

from textual.app import ComposeResult
from textual.reactive import var
from textual.containers import HorizontalGroup, VerticalGroup, VerticalScroll
from textual.widgets import Static, Label, ContentSwitcher, Pretty
from textual.events import Key, DescendantBlur, Click

from asyncio import to_thread
import re
import tempfile
import webbrowser
import base64
from io import BytesIO
from PIL import Image
from rich.text import Text
from typing import Any
from enum import Enum

from .cell import (
    CopyTextArea,
    SplitTextArea,
    Cell,
    StaticBtn,
    COLLAPSED_COLOR,
    EXPANDED_COLOR,
)


class OutputCollapseLabel(Label):
    """Custom label to use as the collapse button for the output of a code cell."""

    collapsed = var(False, init=False)  # keep track of collapse state

    def __init__(self, collapsed: bool = False, id: str = "") -> None:
        super().__init__("\n┃", id=id)
        self.collapsed = collapsed

    def on_click(self) -> None:
        """Toggle the collapsed state on clicks."""
        self.collapsed = not self.collapsed

    def watch_collapsed(self, collapsed: bool) -> None:
        """Watched method to switch the content from the outputs to the collapsed label.

        Args:
            collapsed: updated collapsed state.
        """
        code_cell: CodeCell = self.parent.parent

        if collapsed and len(code_cell.outputs) > 0:
            code_cell.output_switcher.current = "collapsed-output"
            self.styles.color = COLLAPSED_COLOR
        else:
            code_cell.output_switcher.current = "outputs"
            self.styles.color = EXPANDED_COLOR

class ExecStatus(Enum):
    """Status for code cell execution."""
    IDLE=0
    QUEUED=1
    RUNNING=2
    ERROR=3


class RunLabel(Label):
    """Custom label used as button for running/interrupting code cell."""

    status: ExecStatus = var(ExecStatus.IDLE, init=False)
    glyphs = {
        ExecStatus.IDLE: "▶", 
        ExecStatus.ERROR: "[red 50%]▶[/]", 
        ExecStatus.RUNNING: "□",
        ExecStatus.QUEUED: "..",
    }  # glyphs representing running state
    toolips = {
        ExecStatus.IDLE: "Idle: Run", 
        ExecStatus.ERROR: "Error: Run", 
        ExecStatus.RUNNING: "Running: Interrupt",
        ExecStatus.QUEUED: "Queued: Interrupt",
    }

    def __init__(self, id: str = "") -> None:
        super().__init__(self.glyphs[ExecStatus.IDLE], id=id)
        self.tooltip = self.toolips[ExecStatus.IDLE]

    async def on_click(self) -> None:
        """Button to run or interrupt code cell."""
        code_cell: CodeCell = self.parent.parent.parent.parent

        if self.status in [ExecStatus.IDLE, ExecStatus.ERROR]:
            await code_cell.run_cell()
        elif self.status in [ExecStatus.RUNNING, ExecStatus.QUEUED]:
            code_cell.interrupt_cell()

    def watch_status(self, status: ExecStatus) -> None:
        """Watcher method to update the glyph and the tooltip depending on running state.

        Args:
            status: the current execution status for the code cell
        """
        self.update(self.glyphs[status])
        self.tooltip = self.toolips[status]


class CodeArea(SplitTextArea):
    """Widget used for editing code. Inherits from the SplitTextArea."""

    closing_map = {"{": "}", "(": ")", "[": "]", "'": "'", '"': '"'}

    def on_key(self, event: Key) -> None:
        """Key press event handler to close brackets and quotes.

        Args:
            event: Key press event.
        """
        if event.key == "ctrl+r":
            event.stop()
            code_cell: CodeCell = self.parent.parent.parent
            if code_cell.status in [ExecStatus.IDLE, ExecStatus.ERROR]:
                self.run_worker(code_cell.run_cell)
        elif event.character in self.closing_map:
            self.insert(f"{event.character}{self.closing_map[event.character]}")
            self.move_cursor_relative(columns=-1)
            event.prevent_default()
            return

        super().on_key(event)


class OutputJson(HorizontalGroup):
    """Widget for displaying application/json output_type."""

    can_focus = True  # Make the widget focusable

    def __init__(self, data):
        super().__init__()
        self.data = data
        self.output_text = OutputText(str(self.data), id="plain-json")

    def compose(self) -> ComposeResult:
        """Composed of
        - Content switcher (initial=pretty-json)
            - Pretty (id=pretty-json)
            - CopyTextArea (id=plain-json)
        """
        self.switcher = ContentSwitcher(initial="pretty-json")
        with self.switcher:
            yield Pretty(self.data, id="pretty-json")
            yield self.output_text

    def _on_focus(self) -> None:
        """Switch to plain-json when focusing on widget."""
        self.switcher.current = "plain-json"

    def on_descendant_blur(self, event: DescendantBlur) -> None:
        """Switch to pretty-json when bluring away from descendants."""
        self.switcher.current = "pretty-json"

    def _on_blur(self) -> None:
        """Swtich to the pretty-json when bluring away from widget unless new focused widget is
        the plain-json.
        """
        if not self.app.focused or self.app.focused != self.output_text:
            self.switcher.current = "pretty-json"


class OutputText(CopyTextArea):
    """Widget for displaying stream/plain error outputs"""

    read_only = True  # make text area read only

    def _on_focus(self) -> None:
        """Add border when focused."""
        self.styles.border = "solid", "gray"

    def _on_blur(self) -> None:
        """Remove border when focused."""
        self.styles.border = None


class OutputImage(HorizontalGroup):
    """Widget for displaying image/png output for code cells."""

    can_focus = True

    def __init__(self, base64_data: str, metadata: dict[str, int]) -> None:
        super().__init__()
        # image from kernel is returned as a base64 encoded data
        self.base64_data = base64_data
        self.decoded = BytesIO(base64.b64decode(base64_data))
        self.image = Image.open(self.decoded)

        self.display_img_btn = StaticBtn(
            content="🖼 Img", id="display-img-btn"
        ).with_tooltip("Press to display image")

    def compose(self) -> ComposeResult:
        """Composed with:
        - HorizontalGroup
            - ImageStaticBtn (id=display-img-btn)
        """
        yield self.display_img_btn

    def on_click(self, event: Click):
        """Method to display the image when `StaticBtn` is clicked. Called from `StaticBtn` when it
        is clicked.

        Args:
            event: the original click event from the `StaticBtn`.
        """
        if event.widget == self.display_img_btn:
            self.image.show()
            event.stop()


class OutputHTML(HorizontalGroup):
    """Widget for displaying html output for code cells."""

    can_focus = True

    def __init__(self, data: list[str] | str) -> None:
        super().__init__()
        # image from kernel is returned as a base64 encoded data
        if isinstance(data, list):
            self.data = "".join(data)
        else:
            self.data = data

        self.display_img_btn = StaticBtn(
            content="🖼 HTML", id="display-html-btn"
        ).with_tooltip("Press to display image")

    def compose(self) -> ComposeResult:
        """Composed with:
        - HorizontalGroup
            - ImageStaticBtn (id=display-html-btn)
        """
        yield self.display_img_btn

    def on_click(self, event: Click):
        """Method to display html when `StaticBtn` is clicked. Called from `StaticBtn` when it
        is clicked.

        Args:
            event: the original click event from the `StaticBtn`.
        """
        if event.widget == self.display_img_btn:
            with tempfile.NamedTemporaryFile("w", delete=False, suffix=".html") as f:
                f.write(self.data)
                url = "file://" + f.name

            # Open in default browser
            webbrowser.open(url)
            event.stop()


class OutputAnsi(VerticalScroll):
    """Widget for displaying ansi output for code cells."""

    can_focus = True  # make widget focusable

    def __init__(self, ansi_string: list[str] | str) -> None:
        super().__init__()
        if isinstance(ansi_string, list):
            text = "\n".join(ansi_string)
        else:
            text = ansi_string

        self.plain_string = self.remove_ansi(text)  # remove the ansi
        self.pretty_string = Text.from_ansi(
            text
        )  # convert ansi to markup

        self.static_output = Static(
            content=self.pretty_string, id="pretty-output"
        )
        self.text_output = OutputText(
            text=self.plain_string, id="plain-output"
        )

    def compose(self) -> ComposeResult:
        """Composed of
        - HorizontalGroup
            - Content switcher (initial=pretty-output)
                - Pretty (id=pretty-output)
                - CopyTextArea (id=plain-output)
        """
        self.switcher = ContentSwitcher(initial="pretty-output")
        with self.switcher:
            yield self.static_output
            yield self.text_output

    def _on_focus(self) -> None:
        """Switch to plain-output when focusing on widget."""
        self.switcher.current = "plain-output"

    def on_descendant_blur(self, event: DescendantBlur) -> None:
        """Switch to pretty-output when bluring away from descendants."""
        self.switcher.current = "pretty-output"

    def _on_blur(self) -> None:
        """Swtich to the pretty-output when bluring away from widget unless new focused
        widget is the plain-output.
        """
        if not self.app.focused or self.app.focused != self.text_output:
            self.switcher.current = "pretty-output"

    def remove_ansi(self, ansi_escaped_string: str) -> str:
        """Returns the strings with ansi escapes removed using regex. Used to remove color from
        code execution outputs.

        Args:
            ansi_escaped_string: input string containing ansi escapes.

        Returns: string without ansi escapes.
        """
        ansi_escape = re.compile(r"\x1b\[[0-9;]*m")
        return ansi_escape.sub("" , ansi_escaped_string)


class CodeCell(Cell):
    """Widget to contain code cells in a notebook"""

    BINDINGS = [
        ("ctrl+r", "run_cell", "Run Cell"),
    ]
    exec_count: int | None = var(
        None, init=False
    )  # Reactive to keep track of the execution count
    cell_type = "code"

    def __init__(
        self,
        notebook,
        source: str = "",
        outputs: list[dict[str, Any]] = [],
        exec_count: int | None = None,
        metadata: dict[str, Any] = {},
        cell_id: str | None = None,
        language: str = "Python",
    ) -> None:
        super().__init__(notebook, source, language, metadata, cell_id)
        self.outputs: list[dict[str, Any]] = outputs
        self.exec_count = exec_count
        self.switcher = ContentSwitcher(id="collapse-content", initial="text")

        self.run_label = RunLabel(id="run-button")
        self.exec_count_display = Static(f"[{self.exec_count or ' '}]", id="exec-count")

        self.input_text = CodeArea.code_editor(
            self.source,
            language=self._language.lower(),
            id="text",
            soft_wrap=True,
            theme="vscode_dark",
        )

        self.output_collapse_btn = OutputCollapseLabel(
            id="output-collapse-button"
        ).with_tooltip("Collapse Output")

        self.outputs_group = VerticalGroup(id="outputs")
        self.output_switcher = ContentSwitcher(id="collapse-outputs", initial="outputs")

    def compose(self) -> ComposeResult:
        """Compose with:
        - VerticalGroup
            - HorziontalGroup
                - HorizontalGroup (id=code-sidebar)
                    - CollapseLabel (id=collapse-button)
                    - VerticalGroup:
                        - RunLabel (id=run-button)
                        - Static (id=exec-count)
                - ContentSwitcher (id=collapse-content)
                    - CodeArea (id=text)
                    - Static (id=collapsed-display)
            - HorizontalGroup (id=output-section)
                - OutputCollapseLabel (id=output-collapse-button)
                - ContentSwitcher (id=collapse-outputs)
                    - VerticalGroup (id=outputs)
                    - Static (id=collapsed-output)
        """
        with HorizontalGroup():
            with HorizontalGroup(id="code-sidebar"):
                yield self.collapse_btn
                with VerticalGroup():
                    yield self.run_label
                    yield self.exec_count_display
            with self.switcher:
                yield self.input_text
                yield self.collapsed_display

        with HorizontalGroup(id="output-section"):
            yield self.output_collapse_btn
            with self.output_switcher:
                yield self.outputs_group
                yield Static("Outputs are collapsed...", id="collapsed-output")

    def on_mount(self):
        """On mount, toggle the display for the output collapse button if there are outputs and
        display the outputs.
        """
        self.output_collapse_btn.display = len(self.outputs) > 0
        self.call_after_refresh(self.update_outputs, self.outputs)

    def escape(self, event: Key):
        """Event handler to be called when the escape key is pressed."""
        self.focus()
        event.stop()

    def watch_exec_count(self, new: int | None) -> None:
        """Watcher for the execution count to update the value of the Static widget when it changes."""
        self.call_after_refresh(
            lambda: self.exec_count_display.update(f"[{new or ' '}]")
        )

    def action_run_cell(self) -> None:
        """Calls the `run_cell` function."""
        self.run_worker(self.run_cell)

    def action_collapse(self) -> None:
        """Collapse the code cell. If the outputs or the code cell is not collapsed,
        collapse it; otherwise, toggle the collapsed state.
        """
        if not self.collapse_btn.collapsed or not self.output_collapse_btn.collapsed:
            self.collapse_btn.collapsed = True
            self.output_collapse_btn.collapsed = True
        else:
            self.collapse_btn.collapsed = not self.collapse_btn.collapsed
            self.output_collapse_btn.collapsed = not self.output_collapse_btn.collapsed

    async def open(self):
        """Defines what it means to open a code cell. Focus on the input_text widget."""
        self.call_after_refresh(self.input_text.focus)

    @staticmethod
    def from_nb(nb: dict[str, Any], notebook) -> "CodeCell":
        """Static method to generate a `CodeCell` from a json/dict that represent a code cell.

        Args:
            nb: the notebook json/dict format of the code cell.
            notebook: the `Notebook` object the code cell belongs too.

        Returns: `CodeCell` from notebook format.

        Raises:
            AssertionError: if no notebook or bad notebook representation.
        """
        # need to have a notebook object and a notebook format
        assert nb
        assert notebook
        # needs to be a valid notebook representation
        for key in ["cell_type", "execution_count", "metadata", "source", "outputs"]:
            assert key in nb
        assert nb["cell_type"] == "code"

        source = nb["source"]
        if isinstance(source, list):
            # join the strings if the input was a multiline string
            source = "".join(source)

        return CodeCell(
            source=source,
            outputs=nb["outputs"],
            exec_count=nb["execution_count"],
            metadata=nb["metadata"],
            cell_id=nb.get("id", None),
            notebook=notebook,
        )

    def to_nb(self) -> dict[str, Any]:
        """Serialize the `CodeCell` to notebook format. Format for code cell:
            {
                "cell_type" : "code",
                "execution_count": 1, # integer or null
                "metadata" : {
                    "collapsed" : True, # whether the output of the cell is collapsed
                    "autoscroll": False, # any of true, false or "auto"
                },
                "source" : ["some code"],
                "outputs": [{
                    # list of output dicts (described below)
                    "output_type": "stream",
                    ...
                }],
            }

        Returns: serialized code cell representation.
        """
        return {
            "cell_type": "code",
            "execution_count": self.exec_count,
            "id": self._cell_id,
            "metadata": self._metadata,
            "outputs": self.outputs,
            "source": self.input_text.text,
        }

    def create_cell(self, source: str) -> "CodeCell":
        """Returns a `CodeCell` with a source. Used for splitting code cell."""
        return CodeCell(notebook=self.notebook, source=source)

    def clone(self, connect: bool = True) -> "CodeCell":
        """Clone a code cell. Used for cut/paste.

        Args:
            connect: whether to keep the pointers to the next and previous cells.

        Returns: cloned code cell.
        """
        clone = CodeCell(
            notebook=self.notebook,
            source=self.input_text.text,
            outputs=self.outputs,
            exec_count=self.exec_count,
            metadata=self._metadata,
            cell_id=self._cell_id,
            language=self._language,
        )
        if connect:
            clone.next = self.next
            clone.prev = self.prev

        return clone

    async def update_outputs(self, outputs: list[dict[str, Any]]) -> None:
        """Generate the widgets to store the different output types that result from running
        code cell.

        Args:
            outputs: list of serialized outputs.
        """
        try:
            self.outputs_group = self.query_one("#outputs", VerticalGroup)
        except:
            return

        self.output_collapse_btn.display = len(outputs) > 0
        # remove the children widgets first
        await self.outputs_group.remove_children()

        for output in outputs:
            match output["output_type"]:
                case "stream":
                    # join the strings and display them in the `OutputText` widget
                    # {
                    #   "output_type" : "stream",
                    #   "name" : "stdout", # or stderr
                    #   "text" : ["multiline stream text"],
                    # }
                    self.outputs_group.mount(OutputAnsi(output["text"]))
                case "error":
                    # display the errors with the `OutputError` widget
                    # {
                    #   "output_type" : "stream",
                    #   'ename' : str,   # Exception name, as a string
                    #   'evalue' : str,  # Exception value, as a string
                    #   'traceback' : list,
                    # }
                    self.outputs_group.mount(OutputAnsi(output["traceback"]))
                    self.status = ExecStatus.ERROR
                case "execute_result" | "display_data":
                    # the display_data and output_result have different formats
                    # {
                    #   "output_type" : "execute_result" | "display_data",
                    #   "execution_count": 42, # if "execute_result"
                    #   "data" : {
                    #     "text/plain" : ["multiline text data"],
                    #     "image/png": ["base64-encoded-png-data"],
                    #     "application/json": {
                    #       # JSON data is included as-is
                    #       "json": "data",
                    #     },
                    #   },
                    #   "metadata" : {
                    #     "image/png": {
                    #       "width": 640,
                    #       "height": 480,
                    #     },
                    #   },
                    # }
                    for type, data in output["data"].items():
                        match type:
                            case "text/plain":
                                # plain text can also use the `OutputAnsi` widget for display
                                self.outputs_group.mount(OutputAnsi(data))
                            case "application/json":
                                # json is displayed with the `OutputJson` widget
                                self.outputs_group.mount(OutputJson(data))
                            case "image/png":
                                # display the images with the `OutputImage` widget
                                metadata = output["metadata"]
                                self.outputs_group.mount(OutputImage(data, metadata))
                            case "text/html":
                                # display the html douput with the `OutputHTHML` widget
                                self.outputs_group.mount(OutputHTML(data))

        self.refresh()

    @property
    def status(self) -> ExecStatus:
        return self.run_label.status

    @status.setter
    def status(self, status: ExecStatus) -> None:
        self.run_label.status = status

    async def run_cell(self) -> None:
        """Run code in code cell with the kernel in a thread. Update the outputs and the
        execution count for the cell.
        """
        # check if there is a kernel for the notebook
        if not self.notebook._is_kernel_connected():
            return

        self.status = ExecStatus.QUEUED
        # outputs, count= await to_thread(self.notebook.notebook_kernel.run_code, self.input_text.text)
        self.notebook._exec_queue.enqueue(self)
        self.status = ExecStatus.IDLE
        # self.handle_exec_completion(outputs, count)
# 
    def handle_exec_completion(self, outputs: list[dict[str, Any]], execution_count: int) -> None:
        self.exec_count = execution_count
        self.outputs = outputs
        self.call_next(self.update_outputs, outputs)  # update the output cells

    def interrupt_cell(self) -> None:
        """Interrupt kernel when running cell."""

        if not self.notebook._is_kernel_connected():
            return

        self.notebook.interrupt_exec()


from textual.app import ComposeResult
from textual.containers import Container
from textual.events import Key
from pathlib import Path
from .cell import CopyTextArea


extension_to_language: dict[str, str] = {
    ".py": "python",
    ".md": "markdown",
    ".yaml": "yaml",
    ".sh": "bash",
    ".css":"css",
    ".go":"go",
    ".html":"html",
    ".java":"java",
    ".js":"javascript",
    ".json":"json",
    ".rs":"rust",
    ".toml":"toml",
    ".xml":"xml",
}

class FileEditor(CopyTextArea):
    def on_key(self, event: Key) -> None:
        """Handle key press event for saving and saving as.

        Args:
            event: the Key event. 
        """
        file: File = self.parent
        match event.key:
            case "ctrl+s":
                file.save_file(file.path)
                event.stop()
            case "ctrl+w":
                file.save_as_file()
                event.stop()


class File(Container):
    """Widget for rendering and editing contents of non-notebook files."""
    def __init__(self, path: str, id: str, term_app) -> None:
        super().__init__(id=id)

        self.path = Path(path)
        self.language = extension_to_language.get(self.path.suffix, None)
        self.source = ""

        self.input_text = FileEditor.code_editor(
            self.source,
            language=self.language,
            id="file-text",
            soft_wrap=True,
            theme="vscode_dark",
        )

        self.term_app = term_app

    def compose(self) -> ComposeResult:
        """Composed with:
        - Container
            - CopyTextArea (id=file-text)
        """
        yield self.input_text

    def on_mount(self) -> None:
        """Load the file if the path exists. Otherwise, create the file where it is."""
        if self.path.exists():
            self.call_after_refresh(self.load_file)

        self.call_next(self.focus_file)

    def save_as_file(self) -> str:
        """Save notebook as a new file."""

        def check_save_as(path: Path | None) -> None:
            """Callback function to save notebook if save as screen dismisses successfully.

            Args:
                path: string if save as screen dismisses with a file path chosen
                    to save notebook at.
            """
            if path:
                self.path = path
                # notify after successfuly serializing and saving notebook
                self.save_file(path)
                self.notify(f"{self.path} saved!")
                # open the saved notebook
                
                # change the tab name
                self.term_app.change_tab_name(self.id, self.path)

        # push the save as screen with the above callback function
        self.app.push_screen("file_save_as_screen", check_save_as)

    def save_file(self, path: str) -> None:
        """Saves the file to the provided path.

        Args:
            path: file path to save file at.
        """
        self.source = self.input_text.text
        with open(path, "w") as f:
            f.write(self.source)
        self.notify(f"{self.path} saved!")

    def load_file(self) -> None:
        """Load contents of a file from a path to the editor."""
        with open(self.path, "r") as f:
            try:
                # if the json decoding fails then the notebook is bad
                self.source = f.read()
                self.input_text.text = self.source
            except UnicodeDecodeError:
                self.notify(f"Failed to read {self.path}.", severity="error")
                self.term_app.remove_tab(str(self.path))
                return

    def focus_file(self) -> None:
        """Focus on the editor area."""
        self.call_next(self.input_text.focus)


from textual.app import ComposeResult
from textual.widgets import Markdown, Static, ContentSwitcher
from textual.events import Key, MouseDown
from textual.containers import HorizontalGroup

from typing import Any

from .cell import Cell, SplitTextArea


class MarkdownCell(Cell):
    """Widget to contain markdown cells in a notebook"""

    cell_type = "markdown"

    def __init__(
        self,
        notebook,
        source: str = "",
        metadata: dict[str, Any] = {},
        cell_id: str | None = None,
    ) -> None:
        super().__init__(notebook, source, "markdown", metadata, cell_id)
        self.collapse_btn.styles.width = 5
        self.switcher = ContentSwitcher(id="collapse-content", initial="markdown")
        self.collapsed_markdown = Static(
            "Collapsed Markdown...", id="collapsed-display"
        )
        self.input_text = SplitTextArea.code_editor(
            self.source, id="text", language="markdown", show_line_numbers=False
        )
        self.markdown = Markdown(self.source, id="markdown")

    def compose(self) -> ComposeResult:
        """Compose with:
        - VerticalGroup
            - HorziontalGroup
                - CollapseLabela (id=collapse-button)
                - ContentSwitcher (id=collapse-content)
                    - SplitTextArea (id=text)
                    - FocusMarkdown (id=markdown)
                    - Static (id=collapsed-display)
        """
        with HorizontalGroup():
            yield self.collapse_btn
            with self.switcher:
                yield self.collapsed_display
                yield self.input_text
                yield self.markdown

    def on_double_click(self, event: MouseDown) -> None:
        """Double click event handler that switches content to the input text from markdown.

        Args:
            event: MouseDown event.
        """
        if self.switcher.current == "markdown":
            self.switcher.current = "text"
            self.input_text.focus()

    def action_collapse(self) -> None:
        """Toggle the collapsed."""
        self.collapse_btn.collapsed = not self.collapse_btn.collapsed

    def escape(self, event: Key) -> None:
        """Event handler to be called when the escape key is pressed."""
        self.render_markdown()
        self.focus()
        event.stop()

    def render_markdown(self) -> None:
        """Changes content switcher's currently displayed widget to the markdown and
        updates the content input_text."""
        self.source = self.input_text.text
        # if not self.source:
        #     self.markdown.update(PLACEHOLDER)
        # else:
        self.markdown.update(self.source)
        self.switcher.current = "markdown"

    @staticmethod
    def from_nb(nb: dict[str, Any], notebook) -> "MarkdownCell":
        """Static method to generate a `MarkdownCell` from a json/dict that represent a
        makrdown cell.

        Args:
            nb: the notebook json/dict format of the markdown cell.
            notebook: the `Notebook` object the markdown cell belongs too.

        Returns: `MarkdownCell` from notebook format.

        Raises:
            AssertionError: if no notebook or bad notebook representation.
        """
        # need to have a notebook object and a notebook format
        assert nb
        # needs to be a valid notebook representation
        for key in ["cell_type", "metadata", "source"]:
            assert key in nb
        assert nb["cell_type"] == "markdown"

        source = nb["source"]
        if isinstance(source, list):
            # join the strings if the input was a multiline string
            source = "".join(source)

        return MarkdownCell(
            notebook=notebook,
            source=source,
            metadata=nb["metadata"],
            cell_id=nb.get("id"),
        )

    def to_nb(self) -> dict[str, Any]:
        """Serialize the `MarkdownCell` to notebook format. Format for markdown cell:
            {
                "cell_type" : "markdown",
                "metadata" : {},
                "source" : ["some *markdown*"],
            }

        Returns: serialized markdown cell representation.
        """
        return {
            "cell_type": "markdown",
            "metadata": self._metadata,
            "source": self.input_text.text,
            "id": self._cell_id,
        }

    def create_cell(self, source) -> "MarkdownCell":
        """Returns a `MarkdownCell` with a source. Used for splitting code cell."""
        return MarkdownCell(self.notebook, source)

    def clone(self, connect: bool = True) -> "MarkdownCell":
        """Clone a markdown cell. Used for cut/paste.

        Args:
            connect: whether to keep the pointers to the next and previous cells.

        Returns: cloned markdown cell.
        """
        clone = MarkdownCell(
            notebook=self.notebook,
            source=self.input_text.text,
            metadata=self._metadata,
            cell_id=self._cell_id,
        )
        if connect:
            clone.next = self.next
            clone.prev = self.prev
        return clone

    async def open(self):
        """Defines what it means to open a markdown cell. Focus on the input_text widget."""
        self.switcher.current = "text"
        self.call_after_refresh(self.input_text.focus)



from textual.app import App, ComposeResult
from textual.widgets import (
    Footer,
    Header,
    DirectoryTree,
    Tab,
    Tabs,
    ContentSwitcher,
    Label,
    Button,
)
from textual.screen import Screen
from textual.containers import Horizontal, Vertical, Grid
from textual.events import Key

from pathlib import Path
import os.path
from argparse import ArgumentParser

from . import __version__
from .notebook import Notebook, DEFAULT_FILE_NAME
from .save_as_screen import SaveAsScreen
from .file import File


class QuitScreen(Screen):
    """Screen with a dialog to quit."""

    def compose(self) -> ComposeResult:
        """Composed with:
        - Screen
            - Grid (id=dialog)
                - Label (id=question)
                - Button (id=quit)
                - Button (id=cancel)
        """
        yield Grid(
            Label("Are you sure you want to quit?", id="question"),
            Button("Quit", variant="error", id="quit"),
            Button("Cancel", variant="primary", id="cancel"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """Button pressed event handler that quits app or pops screen.

        Args:
            event: button pressed event.
        """
        if event.button.id == "quit":
            self.app.exit()
        else:
            self.app.pop_screen()

    def on_key(self, event: Key) -> None:
        """Key event handler that pops screen when escapa is pressed.

        Args:
            event: key event.
        """
        match event.key:
            case "escape":
                self.app.pop_screen()
                event.stop()


class DirectoryNav(DirectoryTree):
    """Custom directory tree with key navigation."""

    BINDINGS = [
        ("backspace", "back_dir", "Go up a directory"),
    ]
    selected_dir: str | None = None  # keep track of the selected directory

    def action_back_dir(self) -> None:
        """Moves up a directory."""
        parent = Path(self.path).resolve().parent
        self.path = parent

    def on_directory_tree_directory_selected(
        self, event: DirectoryTree.DirectorySelected
    ) -> None:
        """Directory selected event handler that sets the current directory to selected.

        Args:
            event: directory selected event.
        """
        self.path = Path(event.path).resolve()


class Erys(App):
    """A Textual app for editing and running Python notebooks."""

    CSS_PATH = "./styles.tcss"
    # SCREENS = {
    #     "quit_screen": QuitScreen,
    #     "nb_save_as_screen": SaveAsScreen,
    #     "file_save_as_screen": SaveAsScreen,
    # }
    BINDINGS = [
        ("ctrl+n", "new_notebook", "New Notebook"),
        ("ctrl+k", "close", "Close Tab"),
        ("ctrl+l", "clear", "Clear Tabs"),
        ("d", "toggle_directory_tree", "Toggle Directory Tree"),
        ("ctrl+q", "push_screen('quit_screen')", "Quit"),
    ]

    def __init__(self, paths: list[str]) -> None:
        super().__init__()
        self.theme = "textual-dark"
        # check if the provided file paths are python notebooks
        self.paths = [
            os.path.relpath(path, Path.cwd())
            for path in paths
            if (Path(path).exists() or Path(path).parent.is_dir())
        ]
        self.tabs = Tabs(
            *[Tab(path, id=f"tab{idx}") for idx, path in enumerate(self.paths)]
        )
        self.cur_tab = len(paths)
        self.path_to_tab_id: dict[str, int] = {}  # maps from tab id to notebook id
        
        self.dir_tree = DirectoryNav(Path.cwd(), id="file-tree")
        self.switcher = ContentSwitcher(id="tab-content")

    def compose(self) -> ComposeResult:
        """Composed with:
        - App
            - Header
            - Horizontal
                - DirectoryNav (id=file-tree)
                - Vertical
                    - Tabs
                    - ContentSwitcher (id=tab-content)
                        - Notebook
            - Footer
        """

        yield Header(show_clock=True, time_format="%I:%M:%S %p")

        with Horizontal():
            yield self.dir_tree

            with Vertical():
                yield self.tabs
                with self.switcher:
                    for idx, path in enumerate(self.paths):
                        self.path_to_tab_id[path] = f"tab{idx}"
                        if Path(path).suffix == ".ipynb":
                            yield Notebook(path, f"tab{idx}", self)
                        else:
                            yield File(path, f"tab{idx}", self)

        yield Footer()

    def on_mount(self) -> None:
        """Mount event handler that creates a new notebook if none are opened when app starts,
        then focuses on the tabs.
        """
        self.install_screen(QuitScreen, "quit_screen")
        self.install_screen(lambda: SaveAsScreen(is_notebook=True), "nb_save_as_screen")
        self.install_screen(lambda: SaveAsScreen(is_notebook=False), "file_save_as_screen")

        if len(self.paths) == 0:
            self.dir_tree.display = True
            self.dir_tree.focus()
        else:
            self.dir_tree.display = False
            self.tabs.active = f"tab{self.cur_tab - 1}"

    def on_tabs_tab_activated(self, event: Tabs.TabActivated) -> None:
        """Tab activated event handler that switches content to show notebook the tab belongs to.

        Args:
            event: tab activated event.
        """
        if event.tab is None or str(event.tab.label) not in self.path_to_tab_id:
            pass
        else:
            file_id = self.path_to_tab_id[str(event.tab.label)]
            self.switcher.current = f"{file_id}"

    def on_directory_tree_file_selected(
        self, event: DirectoryTree.FileSelected
    ) -> None:
        """File selected event handler that opens notebook or file.

        Args:
            event: file selected event.
        """
        self.open_file(event.path)

    def on_key(self, event: Key) -> None:
        """Key event handler that
            - moves focus to the Tabs when escape is pressed
            - moves focus to latest Notebook when enter is pressed (and tabs was focused on)

        Args:
            event: key event.
        """
        match event.key:
            case "escape":
                self.set_focus(self.tabs)
            case "enter":
                if not self.switcher.current:
                    return

                # if currently focused on the tabs, change focus to the latest notebook
                if isinstance(self.app.focused, Tabs):
                    file = self.switcher.query_one(f"#{self.switcher.current}", Notebook | File)
                    file.focus_file()

    def action_toggle_directory_tree(self) -> None:
        """Toggle whether the directory tree is displayed."""
        self.dir_tree.display = not self.dir_tree.display
        if self.dir_tree.display:
            self.set_focus(self.dir_tree)
        elif cur_file := self.switcher.current:
            file = self.switcher.query_one(f"#{cur_file}", Notebook | File)
            file.focus_file()
        else:
            self.set_focus(self.tabs)

    def action_new_notebook(self) -> None:
        """Create a new notebook with the path set to `DEFAULT_FILE_NAME{self.cur_tab}`."""
        # for a new notebook the notebook id is the same as the tab id
        tab_id = f"tab{self.cur_tab}"
        # add a new tab
        self.notebook_path = f"{DEFAULT_FILE_NAME}{self.cur_tab}"
        self.tabs.add_tab(Tab(self.notebook_path, id=tab_id))
        self.path_to_tab_id[self.notebook_path] = tab_id

        new_notebook = Notebook(self.notebook_path, tab_id, self)
        self.switcher.mount(new_notebook)

        # set the new tab to be the active one
        # no need to change what is the current notebook since that will be set by the
        # tab activated event handler
        self.tabs.active = tab_id
        self.cur_tab += 1

    def action_close(self) -> None:
        """Remove active tab."""
        active_tab = self.tabs.active_tab
        self.remove_tab(active_tab.label)

    def action_clear(self) -> None:
        """Clear the tabs."""
        self.tabs.clear()
        for child in self.switcher.children:
            child.remove()

        # clear the map and the switcher's currently displayed widget to avoid errors
        self.path_to_tab_id = {}
        self.switcher.current = None

    def change_tab_name(self, tab_id: str, new_path: str) -> None:
        """Update the key in the map from path to tab id when a saved is saved as new.
        
        Args:
            tab_id: the tab id for the notebook with the new path.
            new_path: the new path for the file.
        """
        if tab_id not in self.path_to_tab_id:
            pass

        path = os.path.relpath(new_path, Path.cwd())
        target_tab: Tab = self.tabs.query_one(f"#{tab_id}", Tab)

        del self.path_to_tab_id[target_tab.label]

        target_tab.update(path)
        self.path_to_tab_id[path] = tab_id
        
        # reload the directory tree
        self.dir_tree.path = self.dir_tree.path

    def remove_tab(self, nb_path: str) -> None:
        """Deletes tab belonging to notebook with path `nb_path`.

        Args:
            nb_path: path of the notebook whose tab is being removed.
        """
        # if a tab is active then remove it and the notebook
        tab_id = self.path_to_tab_id[nb_path]
        target_tab: Tab = self.tabs.query_one(f"#{tab_id}", Tab)

        if target_tab is not None:
            self.tabs.remove_tab(target_tab.id)
            self.switcher.remove_children(f"#{tab_id}")
            del self.path_to_tab_id[target_tab.label]

        # set the switcher's currently displayed widget to none to avoid errors
        if len(self.path_to_tab_id) == 0:
            self.switcher.current = None

    def open_file(self, path: Path) -> None:
        """Open a file. Either change tabs to the file if it is already loaded or 
        create a new file (notebook or regular) in a new tab and swtich to that tab.
        
        Args:
            path: path to the file.
        """
        path = os.path.relpath(path, Path.cwd())

        if path in self.path_to_tab_id:
            tab_id = self.path_to_tab_id[path]
            self.tabs.active = tab_id
            return

        tab_id = f"tab{self.cur_tab}"
        self.tabs.add_tab(Tab(path, id=tab_id))
        self.path_to_tab_id[path] = tab_id

        if Path(path).suffix == ".ipynb":
            new_file = Notebook(path, tab_id, self)
        else:
            new_file = File(path, tab_id, self)

        self.switcher.mount(new_file)

        self.tabs.active = tab_id
        self.cur_tab += 1

def main():
    parser = ArgumentParser(
        "erys",
        description="Terminal Interface for Jupyter Notebooks."
    )

    parser.add_argument("notebooks", nargs="*", help="One or more notebooks to open", type=Path)
    parser.add_argument("--version", action="version", version=__version__)
    args = parser.parse_args()

    app = Erys(args.notebooks)
    app.run()


if __name__ == "__main__":
    main()
